var Validator = function (schema, object) {
    var typeDefinitions = [
        'boolean',
        'number',
        'string',
        'object',
        'array',
        'date',
        'any'
    ];

    var keysSpecified = [
        'optional',
        'type',
        'test',
        'minLength',
        'maxLength',
        'minValue',
        'maxValue',
        'defaultValue',
        'allowedValues'
    ];

    var compared = {
        status: 'success',
        value: {}
    };

    var addError = function (key, message) {
        compared.status = 'error';
        if (!compared.errors) {
            compared.errors = [];
        }
        compared.errors.push({
            key: key,
            message: message
        });
    };

    var checkTypeExists = function (type) {
        var wrongType = false;
        var types = type.split('|');

        for (var j = 0; j < types.length; j++) {
            var itemType = types[j];

            if (!~typeDefinitions.indexOf(itemType)) {
                wrongType = true;
                break;
            }
        }

        if (!wrongType) {
            return true;
        } else {
            throw new Error('Wrong type definition in schema');
        }
    };

    var checkType = function (value) {
        var type = typeof value;
        if (type === 'object') {
            type = Array.isArray(value) ? 'array' : type;
        }

        if (type === 'object') {
            type = new Date(value) != 'Invalid Date' ? 'date' : type;
        }
        return type;
    };

    var validateObj = function (obj) {
        // need to check first level elements first
        var keys = Object.keys(schema);
        for (var i = 0; i < keys.length; i++) {
            var key = keys[i];
            var schemaItem = schema[key];
            checkTypeExists(schemaItem.type);

            var item = obj[key];
            var optional = schemaItem.optional || false;
            if (schemaItem.defaultValue && !optional) {
                throw new Error('Default value is uses only in optional field');
            }

            // check optional
            if (!(key in obj) && !optional) {
                addError(key, 'value is required');
                continue;
            } else if (!(key in obj)) {
                if (schemaItem.defaultValue) {
                    compared.value[key] = schemaItem.defaultValue;
                }
                continue;
            }

            // check type
            if (!~schemaItem.type.split('|').indexOf(checkType(item)) && schemaItem.type !== 'any') {
                addError(key, 'wrong type');
                continue;
            }

            // string length
            if ((schemaItem.minLength || schemaItem.maxLength) && checkType(item) === 'string') {
                var min = schemaItem.minLength || 0;
                var max = schemaItem.maxLength || 16777215;
                if (max < min || checkType(min) !== 'number' || checkType(max) !== 'number' || min < 0) {
                    throw new Error('Wrong string length declaration');
                }
                if (item.length < min) {
                    addError(key, 'string length is less than min length');
                    continue;
                } else if (item.length > max) {
                    addError(key, 'string length is bigger than max length');
                    continue;
                }
            }

            // string regexp
            if (schemaItem.test && checkType(item) === 'string') {
                var _regExp;
                try {
                    _regExp = new RegExp(schemaItem.test);
                } catch (err) {
                    throw new Error('Wrong regexp string declaration');
                    continue;
                }
                var matched = item.match(_regExp);
                if (!matched) {
                    addError(key, 'string is not allowed by regular expression');
                    continue;
                } else {
                    item = matched[0];
                }
            }

            // number value
            if ((schemaItem.minValue || schemaItem.maxValue) && checkType(item) === 'number') {
                var min = schemaItem.minValue || 0;
                var max = schemaItem.maxValue || 16777215;
                if (max < min || checkType(min) !== 'number' || checkType(max) !== 'number') {
                    throw new Error('Wrong number value declaration');
                }
                if (item < min) {
                    addError(key, 'number value is less than min');
                    continue;
                } else if (item > max) {
                    addError(key, 'number value is bigger than max');
                    continue;
                }
            }

            // allowed values
            if (schemaItem.allowedValues) {
                if (!~schemaItem.allowedValues.indexOf(item)) {
                    addError(key, 'value is not in list of allowed');
                    continue;
                }
            }

            compared.value[key] = item;
        }
    };

    var validate = function (obj) {
        try {
            if (Array.isArray(obj) || typeof obj !== 'object') {
                throw new Error('Please specify object to validate');
            }
            validateObj(obj);
        } catch (error) {
            compared.status = 'schemaError';
            compared.message = error.message;
        }

        if (compared.status !== 'success') {
            delete compared.value;
        }

        return compared;
    };

    return validate(object);
};

module.exports = Validator;
